package space.blokk.net

import io.netty.buffer.ByteBuf
import space.blokk.chat.TextComponent
import space.blokk.event.EventBusWrapper
import space.blokk.net.packet.OutgoingPacket
import space.blokk.net.packet.Protocol
import space.blokk.player.GameMode
import space.blokk.player.Player
import space.blokk.world.WorldAndLocationWithRotation
import java.net.InetAddress
import java.util.*
import kotlin.coroutines.CoroutineContext

abstract class Session {
    @Suppress("LeakingThis")
    val events = EventBusWrapper<Session>(this)

    /**
     * The IP address of this session
     */
    abstract val address: InetAddress

    /**
     * Unconfined [CoroutineContext] which is cancelled when the session is disconnected.
     */
    abstract val coroutineContext: CoroutineContext

    /**
     * The brand name the client optionally sent during the login procedure.
     * [ClientBrandReceivedEvent][space.blokk.net.event.ClientBrandReceivedEvent] is emitted when this value changes.
     */
    abstract val brand: String?

    /**
     * The round trip time (in milliseconds) of the KeepAlive packet which is sent to the client every few seconds.
     *
     * It is -1 until the first IncomingKeepAlive packet is received.
     */
    abstract val ping: Int

    /**
     * The player corresponding to this session.
     */
    val player: Player? get() = (state as? State.WithPlayer)?.player

    /**
     * The current state of this session.
     */
    abstract val state: State

    sealed class State {
        object WaitingForHandshake : State()

        object WaitingForStatusRequest : State()

        object WaitingForLoginStart : State()
        class WaitingForEncryptionVerification(val username: String, val verifyToken: ByteArray) : State()
        class Encrypted(val username: String) : State()

        class LoginSucceeded(val username: String, val uuid: UUID) : State()

        class WaitingForClientSettings(
            val username: String,
            val uuid: UUID,
            val flying: Boolean,
            val canFly: Boolean,
            val flyingSpeed: Float,
            val fieldOfView: Float,
            val gameMode: GameMode,
            val initialWorldAndLocation: WorldAndLocationWithRotation,
            val invulnerable: Boolean,
            val reducedDebugInfo: Boolean,
            val selectedHotbarSlot: Int
        ) : State()

        interface WithPlayer {
            val player: Player
        }

        class Joining(override val player: Player) : State(), WithPlayer
        class Playing(override val player: Player) : State(), WithPlayer

        object Disconnected : State()

        inline fun <reified T : State> getOrFail(): T =
            if (this is T) this
            else throw IllegalStateException("Session.state must be ${T::class.java.simpleName}")
    }

    /**
     * The protocol this session is currently using. If this is null, the client is not connected anymore.
     */
    abstract val currentProtocol: Protocol?

    /**
     * Closes the connection with the client.
     *
     * @param reason The message shown to the player. Only used if [currentProtocol] is `LOGIN` or `PLAY`.
     * @param internalReason A short, loggable representation of [reason] only shown to the player in development mode.
     */
    abstract suspend fun disconnect(
        reason: TextComponent = TextComponent of "Disconnected.",
        internalReason: String? = null
    )

    /**
     * Sends a packet.
     */
    abstract suspend fun send(packet: OutgoingPacket)

    /**
     * Sends a plugin message packet.
     */
    abstract suspend fun sendPluginMessage(channel: String, data: ByteBuf)
}
