package space.blokk.player

import kotlinx.coroutines.CoroutineScope
import space.blokk.chat.TextComponent
import space.blokk.net.Session
import space.blokk.player.event.PlayerEvent
import space.blokk.world.Chunk
import space.blokk.world.LocationWithRotation
import space.blokk.world.VoxelLocation
import space.blokk.world.World
import java.util.*
import kotlin.coroutines.CoroutineContext

/**
 * A **real** player.
 */
interface Player {
    val coroutineContext: CoroutineContext get() = session.coroutineContext

    /**
     * The session of this player.
     */
    val session: Session

    /**
     * The name of this player.
     */
    val name: String

    /**
     * The [UUID] of this player.
     */
    val uuid: UUID

    /**
     * The [GameMode] of this player.
     */
    var gameMode: GameMode

    /**
     * The settings this client sent in the `ClientSettings` packet.
     */
    val settings: Settings

    data class Settings(
        val viewDistance: Byte,
        val locale: String,
        val mainHand: Hand,
        val enabledSkinParts: SkinPartsConfiguration,
        val chatColorsEnabled: Boolean,
        val chatMode: ChatMode
    )

    /**
     * The current location of this player.
     */
    var location: LocationWithRotation

    /**
     * The world which currently contains this player.
     */
    val world: World

    /**
     * The index of the hotbar slot which is currently selected.
     * Must be in `0..8`.
     */
    var selectedHotbarSlot: Int

    /**
     * Corresponds to the `reducedDebugInfo` gamerule in vanilla.
     *
     * @see <a href="https://minecraft.gamepedia.com/Debug_screen#Legend">https://minecraft.gamepedia.com/Debug_screen#Legend</a>
     */
    var reducedDebugInfo: Boolean

    /**
     * Whether the player cannot take damage through other entities or the world.
     */
    var invulnerable: Boolean

    /**
     * Whether the player can start flying by itself.
     */
    var canFly: Boolean

    /**
     * Whether the player is currently flying.
     */
    var flying: Boolean

    /**
     * The default is [DEFAULT_FLYING_SPEED].
     */
    var flyingSpeed: Float

    /**
     * Modifies the field of view, like a speed potion.
     *
     * The default is [DEFAULT_FIELD_OF_VIEW].
     */
    var fieldOfView: Float

    var playerListName: TextComponent?
    var compassTarget: VoxelLocation

    val currentlyViewedChunks: List<Chunk>

    companion object {
        const val DEFAULT_FIELD_OF_VIEW = 0.1f
        const val DEFAULT_FLYING_SPEED = 0.05f
    }
}
