package space.blokk.world

import space.blokk.CoordinatePart
import space.blokk.Vector
import kotlin.math.roundToInt

sealed class AbstractWorldAndLocation<T : AbstractLocation> {
    abstract val world: World
    abstract val location: T
}

data class WorldAndLocation(
    override val world: World,
    override val location: Location
) : AbstractWorldAndLocation<Location>()

data class WorldAndLocationWithRotation(
    override val world: World,
    override val location: LocationWithRotation
) : AbstractWorldAndLocation<LocationWithRotation>()

sealed class AbstractLocation {
    abstract val x: Double
    abstract val y: Double
    abstract val z: Double

    /**
     * Converts this Location to a VoxelLocation by converting [x], [y] and [z] to integers using [Double.toInt],
     * in contrast to [roundToBlock] which uses [Double.roundToInt].
     */
    fun toVoxelLocation(): VoxelLocation = VoxelLocation(x.toInt(), y.toInt().toUByte(), z.toInt())

    /**
     * Converts this Location to a VoxelLocation by converting [x], [y] and [z] to integers using [Double.roundToInt],
     * in contrast to [toVoxelLocation] which uses [Double.toInt].
     */
    fun roundToBlock(): VoxelLocation = VoxelLocation(x.roundToInt(), y.roundToInt().toUByte(), z.roundToInt())

    /**
     * Returns a LocationWithRotation composed of this location, [yaw] and [pitch].
     */
    fun withRotation(yaw: Float, pitch: Float) = LocationWithRotation(x, y, z, yaw, pitch)

    /**
     * Returns a pair of [world] and this location.
     */
    abstract infix fun inside(world: World): AbstractWorldAndLocation<*>

    operator fun get(part: CoordinatePart): Double = when (part) {
        CoordinatePart.X -> x
        CoordinatePart.Y -> y
        CoordinatePart.Z -> z
    }
}

/**
 * Represents a set of x, y and z coordinates.
 *
 * Whenever used as a type, [AbstractLocation] should be preferred so that [LocationWithRotation] is also a valid value.
 */
data class Location(override val x: Double, override val y: Double, override val z: Double) : AbstractLocation() {
    override fun inside(world: World) = WorldAndLocation(world, this)

    fun asVector() = Vector(x, y, z)
}

data class LocationWithRotation(
    override val x: Double,
    override val y: Double,
    override val z: Double,
    val yaw: Float,
    val pitch: Float
) : AbstractLocation() {
    override fun inside(world: World) = WorldAndLocationWithRotation(world, this)

    fun toVector() = Vector(x, y, z)
}
