package space.blokk.world

import kotlinx.coroutines.*
import space.blokk.Blokk
import space.blokk.entity.Entity
import space.blokk.util.newSingleThreadDispatcher
import space.blokk.util.supervisorChild
import java.util.*
import kotlin.coroutines.CoroutineContext

/**
 * A Minecraft world.
 */
abstract class World(val uuid: UUID) {
    private val identifier = "World($uuid)"
    private val threadExecutor = newSingleThreadDispatcher(identifier)

    /**
     * [CoroutineContext] confined to the world thread.
     *
     * Is cancelled when the world is unloaded.
     */
    val coroutineContext: CoroutineContext = Blokk.coroutineContext.supervisorChild(identifier) + threadExecutor

    abstract val dimension: Dimension
    abstract val loadedChunks: Map<Chunk.Key, Chunk>
    abstract val isFlat: Boolean

    /**
     * This can be any value.
     *
     * Only the first 8 bytes of the SHA-256 hash of this value are sent to the client
     * (and it seems not to use it in any way).
     * A random value will be used if this is null.
     */
    open val seed: Long? = null

    /**
     * All entities in this world. Use [spawnEntity] for spawning new ones.
     */
    abstract val entities: Collection<Entity>

    abstract fun getChunk(key: Chunk.Key): Chunk

    /**
     * Returns the chunk containing the voxel at [location].
     */
    fun getChunk(location: VoxelLocation): Chunk = getChunk(Chunk.Key.from(location))

    /**
     * Returns the voxel at [location].
     */
    fun getVoxel(location: VoxelLocation): Voxel = getChunk(location).getVoxel(location)

    /**
     * Returns all voxels in the cube with the corner points [cornerA] and [cornerB].
     *
     * @param hollow Whether the cube is hollow
     */
    @OptIn(ExperimentalStdlibApi::class)
    fun getVoxelsInCube(cornerA: VoxelLocation, cornerB: VoxelLocation, hollow: Boolean = false): List<Voxel> {
        val xList = if (hollow) listOf(cornerA.x, cornerB.x).distinct() else (cornerA.x..cornerB.x).toList()
        val zList = if (hollow) listOf(cornerA.z, cornerB.z).distinct() else (cornerA.z..cornerB.z).toList()

        val yList =
            if (hollow) listOf(cornerA.y.toInt(), cornerB.y.toInt()).distinct()
            else (cornerA.y.toInt()..cornerB.y.toInt()).toList()

        return buildList {
            for(x in xList) {
                for(y in yList) {
                    for(z in zList) {
                        add(getVoxel(VoxelLocation(x, y.toUByte(), z)))
                    }
                }
            }
        }
    }

    /**
     * Returns all voxels in a sphere with the specified [center] and [radius].
     */
    fun getVoxelsInSphere(
        center: VoxelLocation,
        radius: Int,
        hollow: Boolean
    ): List<Voxel> {
        // Implementation example: https://git.io/JLysq
        TODO()
    }

    /**
     * Spawns [entity][Entity] in this world.
     */
    abstract fun spawnEntity(entity: Entity)

    suspend fun destroy() = withContext(coroutineContext) {
        // TODO: Move or kick players
        coroutineContext.cancel()

        coroutineScope {
            loadedChunks.values.forEach { launch { (it as? Chunk.Unloadable)?.unload() } }
        }

        threadExecutor.close()
    }
}
