package space.blokk

import kotlinx.coroutines.CancellableContinuation
import kotlinx.coroutines.runBlocking
import kotlinx.coroutines.suspendCancellableCoroutine
import space.blokk.logging.Logger
import space.blokk.server.Server
import java.util.*
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.Executors
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.TimeUnit
import kotlin.collections.LinkedHashSet
import kotlin.coroutines.resume

/**
 * Basically ExecutorService but for coroutines and with ticks.
 */
class BlokkScheduler : Scheduler {
    private lateinit var executor: ScheduledExecutorService
    private val tasks = ConcurrentHashMap.newKeySet<Task<out Any>>()
    private val shutdownTasks = Collections.synchronizedSet(LinkedHashSet<suspend () -> Unit>())

    val logger = Logger("Scheduler", false)

    inner class Task<R : Any>(
        val fn: suspend () -> R,
        val interval: Int?,
        var ticksUntilExecution: Int
    ) : Scheduler.Task {
        @Volatile
        var cancelled: Boolean = false

        override fun cancel() {
            tasks.remove(this)
            cancelled = true
        }
    }

    fun startTicking() {
        if (this::executor.isInitialized) throw IllegalStateException("Ticking was already started")
        val interval = 1000L / Server.TICKS_PER_SECOND

        executor = Executors.newSingleThreadScheduledExecutor { r -> Thread(r, "Scheduler") }
        executor.scheduleAtFixedRate({
            runBlocking {
                val startTime = System.currentTimeMillis()

                for (task in tasks) {
                    if (!task.cancelled) {
                        if (task.ticksUntilExecution <= 0) {
                            try {
                                task.fn()

                                if (task.interval != null) {
                                    task.ticksUntilExecution = task.interval
                                } else tasks.remove(task)
                            } catch (e: Exception) {
                                tasks.remove(task)
                                e.printStackTrace()
                            }
                        } else {
                            task.ticksUntilExecution -= 1
                        }
                    }
                }

                val duration = System.currentTimeMillis() - startTime
                if (duration > interval) logger.warn("Last tick took ${duration}ms, but should only take ${interval}ms")
            }
        }, 0, interval, TimeUnit.MILLISECONDS)
    }

    private fun stopTicking() {
        if (!this::executor.isInitialized) throw IllegalStateException("Ticking was not started")

        executor.shutdown()
        executor.awaitTermination(3, TimeUnit.SECONDS)
    }

    suspend fun shutdown() {
        stopTicking()
        shutdownTasks.forEach { it.invoke() }
    }

    override fun executeAfter(delay: Int, block: suspend () -> Unit): Scheduler.Task {
        val task = Task(block, null, delay)
        tasks.add(task)
        return task
    }

    override fun executeRepeating(interval: Int, delay: Int, block: suspend () -> Any): Scheduler.Task {
        val task = Task(block, interval, delay)
        tasks.add(task)
        return task
    }

    override fun registerShutdownTask(block: suspend () -> Unit): Scheduler.Task {
        shutdownTasks.add(block)

        return object : Scheduler.Task {
            override fun cancel() {
                shutdownTasks.remove(block)
            }
        }
    }

    override suspend fun <R : Any> runAfter(delay: Int, block: suspend () -> R): R {
        lateinit var continuation: CancellableContinuation<R>

        val fn = suspend {
            continuation.resume(block())
        }

        val task = Task(fn, null, delay)

        return suspendCancellableCoroutine {
            it.invokeOnCancellation {
                task.cancel()
            }

            continuation = it
            tasks.add(task)
        }
    }
}
