package space.blokk

import com.sksamuel.hoplite.ConfigFilePropertySource
import com.sksamuel.hoplite.ConfigLoader
import com.sksamuel.hoplite.ConfigSource
import kotlinx.coroutines.*
import space.blokk.command.Command
import space.blokk.config.BlokkConfig
import space.blokk.event.BlokkEventBus
import space.blokk.event.BlokkEventHandlerPositionManager
import space.blokk.event.EventBus
import space.blokk.event.EventHandlerPositionManager
import space.blokk.logging.BlokkLoggingOutputProvider
import space.blokk.logging.Logger
import space.blokk.net.BlokkSocketServer
import space.blokk.player.Player
import space.blokk.plugin.BlokkPluginManager
import space.blokk.recipe.Recipe
import space.blokk.server.Server
import space.blokk.server.event.ServerEvent
import space.blokk.util.EncryptionUtils
import space.blokk.world.BiomeRegistry
import space.blokk.world.Dimension
import java.io.File
import java.security.KeyPair
import java.util.concurrent.Executors
import kotlin.coroutines.CoroutineContext
import kotlin.system.exitProcess

// TODO: Consider using DI because this improves testability

class BlokkServer internal constructor() : Server {
    val logger = Logger("Server")
    private val socketServer = BlokkSocketServer(this)

    val keyPair: KeyPair =
        try {
            EncryptionUtils.generateKeyPair()
        } catch (e: Exception) {
            failInitialization(e)
        }

    val x509EncodedPublicKey: ByteArray = EncryptionUtils.generateX509Key(keyPair.public).encoded

    override val coroutineContext: CoroutineContext =
        CoroutineName("Server") + SupervisorJob() +
                Executors.newSingleThreadExecutor { r -> Thread(r, "server") }.asCoroutineDispatcher()

    init {
        coroutineContext.job.invokeOnCompletion {
            println(it)
        }
    }

    override val sessions by socketServer::sessions
    override val players get() = sessions.mapNotNull { it.player }

    override val pluginManager = BlokkPluginManager(this)
    override val serverDirectory: File =
        File(BlokkServer::class.java.protectionDomain.codeSource.location.toURI()).let {
            it.mkdirs()
            if (VERSION == "development") it.resolve("../../../../../serverData").normalize()
            else it
        }

    override val dimensionRegistry = Registry<Dimension>()
    override val recipeRegistry = Registry<Recipe>()
    override val commandRegistry = Registry<Command>()
    override val biomeRegistry = BiomeRegistry()

    override val loggingOutputProvider = BlokkLoggingOutputProvider
    override val scheduler = BlokkScheduler()

    val config = ConfigLoader.Builder()
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(serverDirectory.resolve("./config.yml").normalize()),
                optional = true
            )
        )
        .addPropertySource(
            ConfigFilePropertySource(
                ConfigSource.FileSource(File(javaClass.classLoader.getResource("default-config.yml")!!.toURI()))
            )
        )
        .build().loadConfigOrThrow<BlokkConfig>()

    override val minimumLogLevel = config.minLogLevel
    override val developmentMode: Boolean = config.developmentMode

    override val eventBus = BlokkEventBus(developmentMode)
    override val eventHandlerPositions = BlokkEventHandlerPositionManager()

    private fun failInitialization(t: Throwable): Nothing {
        logger.error("Server initialization failed:", t)
        exitProcess(1)
    }

    fun start() {
        logger info "Starting BlokkServer ($VERSION_WITH_V)"
        logger trace "Configuration: $config"

        pluginManager.loadAll()
        pluginManager.enableAll()

        socketServer.bind()
        logger info "Listening on ${config.host}:${config.port}"

        scheduler.startTicking()
    }

    override fun shutdown() {
        runBlocking {
            scheduler.shutdown()
        }
    }

    companion object {
        val VERSION = BlokkServer::class.java.`package`.implementationVersion ?: "development"
        val VERSION_WITH_V = if (VERSION == "development") VERSION else "v$VERSION"

        @JvmStatic
        fun main(args: Array<String>) {
            BlokkServer().also { setServerInstance(it) }.start()
        }

        fun setServerInstance(instance: Server) {
            val clazz = Class.forName("space.blokk.BlokkKt")
            val field = clazz.getDeclaredField("serverInstance")
            field.isAccessible = true
            field.set(null, instance)
            field.isAccessible = false
        }
    }
}
