package space.blokk.net

import io.netty.buffer.ByteBuf
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import space.blokk.BlokkServer
import space.blokk.chat.ChatColor
import space.blokk.chat.ChatComponent
import space.blokk.chat.TextComponent
import space.blokk.event.*
import space.blokk.logging.Logger
import space.blokk.net.MinecraftProtocolDataTypes.readVarInt
import space.blokk.net.event.PacketReceivedEvent
import space.blokk.net.event.PacketSendEvent
import space.blokk.net.packet.IncomingPacket
import space.blokk.net.packet.OutgoingPacket
import space.blokk.net.packet.Protocol
import space.blokk.net.packet.handshaking.HandshakingProtocol
import space.blokk.net.packet.login.LoginProtocol
import space.blokk.net.packet.play.OutgoingKeepAlivePacket
import space.blokk.net.packet.play.OutgoingPluginMessagePacket
import space.blokk.net.packet.play.PlayProtocol
import space.blokk.net.packet.status.StatusProtocol
import space.blokk.server.event.SessionInitializedEvent
import space.blokk.util.awaitSuspending
import space.blokk.util.supervisorChild
import java.io.IOException
import java.net.InetAddress
import java.net.InetSocketAddress
import javax.crypto.SecretKey
import kotlin.coroutines.CoroutineContext
import kotlin.properties.Delegates

class BlokkSession(private val channel: io.netty.channel.Channel, val server: BlokkServer) : Session() {
    override val address: InetAddress = (channel.remoteAddress() as InetSocketAddress).address

    private val identifier = "BlokkSession(${address.hostAddress})" + System.identityHashCode(this)
    val logger = Logger(identifier)

    override val coroutineContext: CoroutineContext = server.coroutineContext.supervisorChild(identifier)
    val scope = CoroutineScope(coroutineContext)

    override var brand: String? = null
    override var ping: Int = -1

    override var state: State = State.WaitingForHandshake

    override val currentProtocol: Protocol?
        get() = when (state) {
            is State.WaitingForHandshake -> HandshakingProtocol

            is State.WaitingForStatusRequest -> StatusProtocol

            is State.WaitingForLoginStart,
            is State.WaitingForEncryptionVerification,
            is State.Encrypted -> LoginProtocol

            is State.LoginSucceeded,
            is State.WaitingForClientSettings,
            is State.Joining,
            is State.Playing -> PlayProtocol

            is State.Disconnected -> null
        }

    private var packetsChannel: Channel<IncomingPacket>? = null
    private val packetDataChannel: Channel<ByteBuf> = Channel(Channel.UNLIMITED)

    fun launchPacketsChannelWorker() {
        val channel = Channel<IncomingPacket>(Channel.UNLIMITED)
        packetsChannel = channel

        scope.launch {
            for (packet in channel) {
                handlePacket(packet)
            }
        }
    }

    private suspend fun handlePacket(packet: IncomingPacket) {
        logger.trace { "Packet received: $packet" }

        server.eventBus.emit(PacketReceivedEvent(this@BlokkSession, packet)).ifNotCancelled {
            SessionPacketReceivedEventHandler.handle(this@BlokkSession, packet)
        }
    }

    init {
        logger trace "Created"

        scope.launch {
            for (data in packetDataChannel) {
                val packetID = data.readVarInt()
                val codec = currentProtocol!!.incomingPacketCodecsByID[packetID]

                if (codec == null) {
                    val message = "Client sent an unknown packet (ID: $packetID)"

                    if (server.config.developmentMode) {
                        logger warn "$message. This will cause the client to disconnect in production mode."
                        continue
                    } else throw IOException(message)
                }

                val packet = codec.decode(data)

                if (packetsChannel != null) {
                    // The protocol will not change anymore, so we can safely decode all packets with the current one
                    if (!packetsChannel!!.offer(packet)) {
                        logger.warn("The packet buffer is full (size: ${server.config.packetsBufferSize})")
                        disconnect(internalReason = "You sent more packets than the packet buffer can handle")
                    }
                } else {
                    // The protocol could change because of this packet, so we must wait
                    handlePacket(packet)
                }
            }
        }
    }

    override suspend fun disconnect(reason: TextComponent, internalReason: String?) {
        if (currentProtocol == LoginProtocol || currentProtocol == PlayProtocol) {
            logger info "Disconnected." + (internalReason?.let { " Internal reason: $it" } ?: "")

            val finalReason = if (server.config.developmentMode && internalReason != null) {
                val additional = listOf<ChatComponent>(
                    TextComponent(
                        "\n\nLogged reason (only shown in development mode): \n",
                        color = ChatColor.GRAY
                    ),
                    TextComponent of internalReason
                )

                reason.copy(extra = reason.extra + additional)
            } else reason

            when (currentProtocol) {
                LoginProtocol -> send(space.blokk.net.packet.login.DisconnectPacket(finalReason))
                PlayProtocol -> send(space.blokk.net.packet.play.DisconnectPacket(finalReason))
            }
        }

        onDisconnect(true)
        channel.close()
    }

    var lastKeepAlivePacketTimestamp by Delegates.notNull<Long>()
    lateinit var keepAliveDisconnectJob: Job

    fun scheduleKeepAlivePacket(isFirst: Boolean = false) {
        scope.launch {
            if (!isFirst) {
                val timeSinceLastPacket = (System.currentTimeMillis() - lastKeepAlivePacketTimestamp).toInt()
                delay(KEEP_ALIVE_PACKET_INTERVAL.toLong() - timeSinceLastPacket)
            }

            lastKeepAlivePacketTimestamp = System.currentTimeMillis()
            send(OutgoingKeepAlivePacket(lastKeepAlivePacketTimestamp))

            keepAliveDisconnectJob = launch {
                delay(server.config.timeout.toMillis())
                disconnect(
                    TextComponent of "Timed out",
                    "The client did not respond to the KeepAlive packet for " +
                            "${server.config.timeout.toMillis()}ms"
                )
            }
        }
    }

    val joinProcedure = LoginAndJoinProcedure(this)

    fun onConnect() = scope.launch {
        logger trace "Connected"
        state = State.WaitingForHandshake
        if (server.eventBus.emit(SessionInitializedEvent(this@BlokkSession)).cancelled) channel.close()
        else server.sessions.add(this@BlokkSession)
    }

    fun onDisconnect(expected: Boolean) {
        if (state == State.Disconnected) return
        if (!expected && currentProtocol != HandshakingProtocol && currentProtocol != StatusProtocol)
            logger trace "The client disconnected unexpectedly"

        val closeException = DisconnectedCancellationException()

        state = State.Disconnected
        server.sessions.remove(this)

        packetDataChannel.close(closeException)
        packetsChannel?.close(closeException)

        coroutineContext.cancel(closeException)
    }

    fun onPacketDataReceived(packetData: ByteBuf) {
        packetDataChannel.offer(packetData)
    }

    fun failBecauseOfClient(message: String) {
        val messageGetter = { "The client caused a connection error: $message" }
        if (server.config.silentNonServerErrors) logger debug messageGetter else logger error messageGetter
        onDisconnect(true)
    }

    fun enableEncryptionCodec(key: SecretKey) {
        logger trace "Enabling encryption codec"
        if (channel.pipeline().get("encryption") != null) return
        channel.pipeline().addBefore("framing", "encryption", EncryptionCodec(this, key))
    }

    fun enableCompressionCodec() {
        logger trace "Enabling compression codec"
        if (channel.pipeline().get("compression") != null) return

        channel.pipeline()
            .addAfter("framing", "compression", CompressionCodec(server.config.packetCompressionThreshold))
    }

    override suspend fun send(packet: OutgoingPacket) {
        if (state is State.Disconnected) throw IllegalStateException("The session is not active anymore")

        logger.trace { "Sending packet: $packet" }
        server.eventBus.emit(PacketSendEvent(this, packet)).ifNotCancelled {
            try {
                channel.writeAndFlush(it.packet).awaitSuspending()
            } catch (t: Throwable) {
                if (!channel.isActive) return
                logger.error("Sending packet failed", t)
            }
        }
    }

    override suspend fun sendPluginMessage(channel: String, data: ByteBuf) {
        if (this.currentProtocol != PlayProtocol) throw IllegalStateException("The session is not using the PLAY protocol")
        send(OutgoingPluginMessagePacket(channel, data))
        data.release()
    }

    companion object {
        const val KEEP_ALIVE_PACKET_INTERVAL = 1000
    }
}
