package space.blokk.event

import kotlinx.coroutines.runBlocking
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import space.blokk.Blokk
import space.blokk.BlokkServer
import strikt.api.expectThat
import strikt.api.expectThrows
import strikt.assertions.isEqualTo
import java.lang.IllegalArgumentException

class EventBusTest {
    class TestEvent: Event() {
        private var count = 0

        fun test(expectedCount: Int) {
            expectThat(count).isEqualTo(expectedCount)
            count++
        }
    }

    object Position1 : EventHandlerPosition()
    object Position2 : EventHandlerPosition()
    object Position3 : EventHandlerPosition()
    object Position4 : EventHandlerPosition()
    object Position5 : EventHandlerPosition()
    object Position6 : EventHandlerPosition()

    @BeforeEach
    fun initServer() {
        // Does not work because Blokk will not be reinitialized. When we use DI, we do no longer need this.
        BlokkServer.setServerInstance(BlokkServer())
    }

    @Test
    fun `handler positions cannot be inserted more than once`() {
        Blokk.eventHandlerPositions.insert(Position1)

        expectThrows<IllegalArgumentException> { Blokk.eventHandlerPositions.insert(Position1) }
        expectThrows<IllegalArgumentException> { Blokk.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1) }
        expectThrows<IllegalArgumentException> { Blokk.eventHandlerPositions.insertBefore(EventHandlerPosition.LAST, Position1) }
    }

    @Test
    fun `handler positions are inserted where they should be`() {
        val expectedOrder = listOf(
            EventHandlerPosition.FIRST,
            Position2,
            Position4,
            Position1,
            Position5,
            EventHandlerPosition.NORMAL,
            Position3,
            Position6,
            EventHandlerPosition.LAST
        )

        val actualOrder = (Blokk.eventHandlerPositions as BlokkEventHandlerPositionManager).positions
        println(actualOrder)

        Blokk.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1, Position5)
        println(actualOrder)
        Blokk.eventHandlerPositions.insertBefore(Position1, Position2, Position4)
        println(actualOrder)
        Blokk.eventHandlerPositions.insert(Position3, Position6)
        println(actualOrder)

        actualOrder.forEachIndexed { index, position ->
            expectThat(index).isEqualTo(expectedOrder.indexOf(position))
        }
    }

    @Test
    fun `handler positions cannot be inserted before the first or after the last`() {
        expectThrows<IllegalArgumentException> {
            Blokk.eventHandlerPositions.insertBefore(EventHandlerPosition.FIRST, Position1)
        }

        expectThrows<IllegalArgumentException> {
            Blokk.eventHandlerPositions.insertAfter(EventHandlerPosition.LAST, Position1)
        }
    }

    @Test
    fun `handlers are invoked in the right order`() {
        val order = listOf(
            EventHandlerPosition.FIRST,
            Position1,
            EventHandlerPosition.NORMAL,
            Position2,
            Position3,
            EventHandlerPosition.LAST
        )

        Blokk.eventHandlerPositions.insertAfter(EventHandlerPosition.FIRST, Position1)
        Blokk.eventHandlerPositions.insertBefore(EventHandlerPosition.NORMAL, Position2)
        Blokk.eventHandlerPositions.insert(Position3)

        order.shuffled().forEach { pos -> Blokk.eventBus.on<TestEvent>(pos) { it.test(order.indexOf(pos)) } }

        runBlocking { Blokk.eventBus.emit(TestEvent()) }
    }
}
